@@ -1,5 +1,7 @@
 import sys
 import math
+import bisect
+from collections import deque
 from PyQt5 import QtCore, QtWidgets, QtGui
 from PyQt5.QtChart import (
     QChart,
@@ -10,7 +12,7 @@ from PyQt5.QtChart import (
     QBarSeries,
     QBarCategoryAxis,
 )
-from PyQt5.QtCore import QPointF
+from PyQt5.QtCore import QPointF, QLineF, Qt
 from PyQt5.QtGui import QPainter, QPainterPath, QColor, QPolygonF
 
 from agents.model import (
@@ -23,6 +25,7 @@ from agents.model import (
     LineChartSpec,
     BarChartSpec,
     HistogramSpec,
+    AgentGraphSpec,
     MonitorSpec,
     EllipseStruct,
     RectStruct,
@@ -52,21 +55,42 @@ class SimulationArea(QtWidgets.QWidget):
         # Default to a black background
         white = QtGui.QColor("black")
         painter.setBrush(white)
-        painter.drawRect(0, 0,
-                         painter.device().width(), painter.device().height())
+        painter.drawRect(
+            0, 0, painter.device().width(), painter.device().height()
+        )
 
         if self.model:
             # Draw tiles
             for tile in self.model.tiles:
                 self.paintTile(painter, tile)
+
             # Draw shapes
             for shape in self.model.get_shapes():
                 c = shape.color
                 painter.setBrush(QtGui.QColor(c[0], c[1], c[2]))
                 if type(shape) is EllipseStruct:
-                    painter.drawEllipse(shape.x,shape.y,shape.w,shape.h)
+                    painter.drawEllipse(shape.x, shape.y, shape.w, shape.h)
                 elif type(shape) is RectStruct:
-                    painter.drawRect(shape.x,shape.y,shape.w,shape.h)
+                    painter.drawRect(shape.x, shape.y, shape.w, shape.h)
+
+            # Draw lines
+            for agent in self.model.agents:
+                painter.setPen(
+                    QColor(agent.color[0], agent.color[1], agent.color[2])
+                )
+                # Pretty terrible code here
+                # Replaces line-tuples in agent.__paths with QLineF's
+                for path in agent.get_paths():
+                    for i in range(len(path)):
+                        if type(path[i]) is tuple:
+                            path[i] = QLineF(
+                                path[i][0][0],
+                                path[i][0][1],
+                                path[i][1][0],
+                                path[i][1][1],
+                            )
+                    painter.drawLines(path)
+
             # Draw agents
             select = None
             for agent in self.model.agents:
@@ -89,7 +113,8 @@ class SimulationArea(QtWidgets.QWidget):
 
     def paintAgent(self, painter, agent):
         r, g, b = agent.color
-        painter.setBrush(QtGui.QColor(r, g, b))
+        painter.setBrush(QColor(r, g, b))
+        painter.setPen(QColor(r, g, b))
         if agent.shape == AgentShape.CIRCLE:
             painter.drawEllipse(
                 agent.x - agent.size / 2,
@@ -180,17 +205,25 @@ class SimulationArea(QtWidgets.QWidget):
         if action == pause_action:
             self.enable_rendering = not self.enable_rendering
 
+
 class QtGraph(QChartView):
     def __init__(self, spec):
         super().__init__(None)
         self.spec = spec
         self.chart = QChart()
-        self.chart.setTitle(str(self.spec.variables))
-        self.chart.legend().hide()
+        # self.chart.setTitle(str(self.spec.variables))
+        # self.chart.legend().hide()
 
-        for color in self.spec.colors:
+        for i in range(len(self.spec.variables)):
             series = QLineSeries()
-            series.setColor(QColor(color[0], color[1], color[2]))
+            series.setColor(
+                QColor(
+                    self.spec.colors[i][0],
+                    self.spec.colors[i][1],
+                    self.spec.colors[i][2],
+                )
+            )
+            series.setName(self.spec.variables[i])
             self.chart.addSeries(series)
 
         self.setMinimumWidth(400)
@@ -202,7 +235,7 @@ class QtGraph(QChartView):
         self.autoscale_y_axis = True
         if self.spec.min_y and self.spec.max_y:
             self.autoscale_y_axis = False
-            self.chart.axes()[1].setRange(self.spec.min_y,self.spec.max_y)
+            self.chart.axes()[1].setRange(self.spec.min_y, self.spec.max_y)
 
         self._updates_per_second = 60
         self._data = []
@@ -222,19 +255,24 @@ class QtGraph(QChartView):
             for i in range(len(self.spec.variables)):
                 data = [datapoint[i] for datapoint in self._data]
                 datapoint = sum(data) / len(data)
-                self.chart.series()[i].append(QPointF(self.chart.series()[i].count() /
-                                                    self._updates_per_second,
-                                                    datapoint))
+                self.chart.series()[i].append(
+                    QPointF(
+                        self.chart.series()[i].count()
+                        / self._updates_per_second,
+                        datapoint,
+                    )
+                )
                 self._min = min(self._min, datapoint)
                 self._max = max(self._max, datapoint)
-        self.chart.axes()[0].setRange(0, (self.chart.series()[0].count() - 1) /
-                                      self._updates_per_second)
+        self.chart.axes()[0].setRange(
+            0, (self.chart.series()[0].count() - 1) / self._updates_per_second
+        )
         diff = self._max - self._min
         if self.autoscale_y_axis:
             if diff > 0:
-                self.chart.axes()[1].setRange(self._min,self._max)
+                self.chart.axes()[1].setRange(self._min, self._max)
             else:
-                self.chart.axes()[1].setRange(self._min-0.5,self._max+0.5)
+                self.chart.axes()[1].setRange(self._min - 0.5, self._max + 0.5)
         self._data = []
 
 
@@ -243,7 +281,7 @@ class QtBarChart(QChartView):
         super().__init__(None)
         self.spec = spec
         self.chart = QChart()
-        self.chart.setTitle(str(self.spec.variables))
+        # self.chart.setTitle(str(self.spec.variables))
         self.chart.legend().hide()
         self.mainset = QBarSet("")
         self.mainset.append([0 for i in range(len(spec.variables))])
@@ -294,7 +332,7 @@ class QtHistogram(QChartView):
         self.chart.legend().hide()
 
         self.mainset = QBarSet("")
-        self.mainset.append([0 for i in range(len(spec.bins))])
+        self.mainset.append([0] * len(spec.bins))
         self.mainset.setColor(
             QColor(spec.color[0], spec.color[1], spec.color[2])
         )
@@ -304,8 +342,21 @@ class QtHistogram(QChartView):
         self.setMinimumWidth(400)
         self.setMinimumHeight(230)
 
+        self.y_ranges = [0, 1, 5, 10, 25, 50, 100, 250, 500, 1000]
+        self.max_y = 1000
+        self.max_y_range = 1000
+        self.lookback = 30
+        self.recent_max_y = deque([self.max_y_range] * self.lookback)
+
+        font = QtGui.QFont()
+        font.setPixelSize(10)
+
         self.axis_x = QBarCategoryAxis()
+        self.axis_x.setLabelsAngle(-90)
+        self.axis_x.setLabelsFont(font)
+
         self.axis_y = QValueAxis()
+        self.axis_y.setRange(0, self.max_y)
         self.axis_x.append(map(str, spec.bins))
         self.chart.addSeries(self.series)
         self.chart.setAxisX(self.axis_x, self.series)
@@ -330,7 +381,111 @@ class QtHistogram(QChartView):
         if len(self._dataset) > 0:
             for i in range(len(self._dataset)):
                 self.mainset.replace(i, self._dataset[i])
-            self.axis_y.setRange(0, max(self._dataset))
+
+            # Calculate max of current values
+            max_y_range = max(self._dataset)
+
+            # Store max value
+            self.recent_max_y.appendleft(max_y_range)
+            if len(self.recent_max_y) > self.lookback:
+                self.recent_max_y.pop()
+
+            # Set max based on the last 30 max values,
+            # to avoid flickering
+            self.max_y_range = max(self.recent_max_y)
+
+            y_range = bisect.bisect_left(self.y_ranges, self.max_y_range)
+            if y_range < len(self.y_ranges):
+                self.max_y = self.y_ranges[y_range]
+            elif max_y_range > self.max_y:
+                self.max_y += self.max_y
+            elif max_y_range < self.max_y / 2:
+                self.max_y = self.max_y / 2
+
+            self.axis_y.setRange(0, self.max_y)
+
+
+class QtAgentGraph(QChartView):
+    def __init__(self, spec):
+        super().__init__(None)
+        self.spec = spec
+        self.chart = QChart()
+        self.chart.setTitle(str(self.spec.variable))
+        self.chart.legend().hide()
+
+        self.setMinimumWidth(400)
+        self.setMinimumHeight(230)
+
+        self.setChart(self.chart)
+        self.setRenderHint(QPainter.Antialiasing)
+        self.chart.createDefaultAxes()
+        self.autoscale_y_axis = True
+        if self.spec.min_y and self.spec.max_y:
+            self.autoscale_y_axis = False
+            self.chart.axes()[1].setRange(self.spec.min_y, self.spec.max_y)
+
+        self.axis_x = QValueAxis()
+        self.axis_y = QValueAxis()
+        self.chart.addAxis(self.axis_x, Qt.AlignBottom)
+        self.chart.addAxis(self.axis_y, Qt.AlignLeft)
+
+        self._updates_per_second = 60
+        self._data = []
+        self._min = 0
+        self._max = 0
+
+    def clear(self):
+        for chart in self.chart.series():
+            chart.clear()
+        self._data = []
+
+    def update_data(self):
+        for a in self.spec.agents:
+            if not hasattr(a, "_agent_series"):
+                a._agent_series = QLineSeries()
+                a._agent_series.setColor(
+                    QColor(a.color[0], a.color[1], a.color[2])
+                )
+                a._agent_series_data = [getattr(a, self.spec.variable)]
+                self.chart.addSeries(a._agent_series)
+                a._agent_series.attachAxis(self.chart.axisX())
+                a._agent_series.attachAxis(self.chart.axisY())
+            else:
+                a._agent_series_data.append(getattr(a, self.spec.variable))
+
+    def redraw(self):
+        for a in self.spec.agents:
+            if hasattr(a, "_agent_series") and len(a._agent_series_data) > 0:
+                datapoint = sum(a._agent_series_data) / len(
+                    a._agent_series_data
+                )
+                a._agent_series.append(
+                    QPointF(
+                        a._agent_series.count() / self._updates_per_second,
+                        datapoint,
+                    )
+                )
+                self._min = min(self._min, datapoint)
+                self._max = max(self._max, datapoint)
+                a._agent_series.setColor(
+                    QColor(a.color[0], a.color[1], a.color[2])
+                )
+            a._agent_series_data = []
+        if len(self.spec.agents) > 0:
+            first_agent = self.spec.agents[0]
+            if hasattr(first_agent, "_agent_series"):
+                first_series = first_agent._agent_series
+                self.chart.axes()[0].setRange(
+                    0, (first_series.count() - 1) / self._updates_per_second
+                )
+                diff = self._max - self._min
+                if self.autoscale_y_axis:
+                    if diff > 0:
+                        self.chart.axes()[1].setRange(self._min, self._max)
+                    else:
+                        self.chart.axes()[1].setRange(
+                            self._min - 0.5, self._max + 0.5
+                        )
 
 
 class ToggleButton(QtWidgets.QPushButton):
@@ -440,7 +595,7 @@ class Application:
 
         # Start timers
         self.logic_timer.start(1000 / 60)
-        self.graphics_timer.start(1000 / 10)
+        self.graphics_timer.start(1000 / 30)
 
     def update_logic(self):
         if not self.model.is_paused():
@@ -491,7 +646,7 @@ class Application:
         checkbox = Checkbox(checkbox_spec.variable)
 
         def update_variable(v):
-            self.model[checkbox_spec.variable] = checkbox.isChecked()
+            setattr(self.model, checkbox_spec.variable, checkbox.isChecked())
 
         checkbox.stateChanged.connect(update_variable)
         row.addWidget(checkbox)
@@ -517,6 +672,11 @@ class Application:
         plots_box.addWidget(histogram)
         self.model.plots.add(histogram)
 
+    def add_agent_graph(self, agent_graph_spec, plots_box):
+        chart = QtAgentGraph(agent_graph_spec)
+        plots_box.addWidget(chart)
+        self.model.plots.add(chart)
+
     def add_controllers(self, rows, controller_box):
         first_row = True
         for row in rows:
@@ -548,6 +708,8 @@ class Application:
                 self.add_bar_chart(plot_spec, plots_box)
             elif type(plot_spec) is HistogramSpec:
                 self.add_histogram(plot_spec, plots_box)
+            elif type(plot_spec) is AgentGraphSpec:
+                self.add_agent_graph(plot_spec, plots_box)
 
 
 def run(model):
